#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <stdbool.h>
#include <unistd.h>
#include <sys/types.h>
#include <fcntl.h>
#include <dlfcn.h>

#include "timer.h"

#include <cmdu.h>
#include <cmdu_ackq.h>
#include <map_module.h>

#define MAP_PLUGIN_LIB	"/usr/lib/libmaputil.so"
#define IEEE1905_LIB	"/usr/lib/libieee1905.so"

int (*t_map_prepare_cmdu_mask)(uint8_t mask[], ...);

int (*t_map_subscribe)(void *bus, void *publisher,
                  const char *name, mapmodule_cmdu_mask_t *mask, void *priv,
                  int (*sub_cb)(void *bus, void *priv, void *data),
                  int (*del_cb)(void *bus, void *priv, void *data),
                  void **subscriber);

int (*t_cmdu_ackq_init)(void *q);
void (*t_cmdu_ackq_free)(void *q);
int (*t_cmdu_ackq_dequeue)(void *q, uint16_t type, uint16_t mid, uint8_t *src,
			     void **cookie);
int (*t_cmdu_ackq_enqueue)(void *q, uint16_t type, uint16_t mid, uint8_t *dest,
			     uint32_t timeout, int resend_cnt, void *cookie);

struct cmdu_buff *(*t_cmdu_alloc_simple)(uint16_t type, uint16_t *mid);


struct cmdu_buff *(*t_cmdu_alloc_custom)(uint16_t type, uint16_t *mid,
				    char *ifname, uint8_t *origin,
				    uint8_t *tlvs, uint32_t tlvslen);

void (*t_cmdu_free)(struct cmdu_buff *c);

uint16_t (*t_cmdu_get_mid)(struct cmdu_buff *c);
uint8_t (*t_cmdu_get_fid)(struct cmdu_buff *c);
uint16_t (*t_cmdu_get_type)(struct cmdu_buff *c);


int (*t_cmdu_parse_tlvs)(struct cmdu_buff *c, struct tlv *tv[][16],
		    struct tlv_policy *policy, int policy_len);


int (*t_cmdu_put_eom)(struct cmdu_buff *c);
int (*t_cmdu_put_tlv)(struct cmdu_buff *c, struct tlv *t);
struct tlv *(*t_cmdu_reserve_tlv)(struct cmdu_buff *c, uint16_t tlv_datalen);
uint16_t (*t_tlv_length)(struct tlv *t);

/* null functions if unresolved */
int u_map_prepare_cmdu_mask(uint8_t mask[], ...)
{
	return -1;
}

int u_map_subscribe(void *bus, void *publisher,
                  const char *name, mapmodule_cmdu_mask_t *mask, void *priv,
                  int (*sub_cb)(void *bus, void *priv, void *data),
                  int (*del_cb)(void *bus, void *priv, void *data),
                  void **subscriber)
{
	return -1;
}

int u_cmdu_ackq_init(void *q)
{
	return -1;
}

void u_cmdu_ackq_free(void *q)
{
}

int u_cmdu_ackq_dequeue(void *q, uint16_t type, uint16_t mid, uint8_t *src,
			void **cookie)
{
	return -1;
}

int u_cmdu_ackq_enqueue(void *q, uint16_t type, uint16_t mid, uint8_t *dest,
			uint32_t timeout, int resend_cnt, void *cookie)
{
	return -1;
}

struct cmdu_buff *u_cmdu_alloc_simple(uint16_t type, uint16_t *mid)
{
	return NULL;
}

struct cmdu_buff *u_cmdu_alloc_custom(uint16_t type, uint16_t *mid,
				    char *ifname, uint8_t *origin,
				    uint8_t *tlvs, uint32_t tlvslen)
{
	return NULL;
}

void u_cmdu_free(struct cmdu_buff *c)
{
}

uint16_t u_cmdu_get_mid(struct cmdu_buff *c)
{
	return 0;
}

uint8_t u_cmdu_get_fid(struct cmdu_buff *c)
{
	return 0;
}

uint16_t u_cmdu_get_type(struct cmdu_buff *c)
{
	return 0xffff;
}

int u_cmdu_parse_tlvs(struct cmdu_buff *c, struct tlv *tv[][16],
		      struct tlv_policy *policy, int policy_len)
{
	return -1;
}

int u_cmdu_put_eom(struct cmdu_buff *c)
{
	return -1;
}

int u_cmdu_put_tlv(struct cmdu_buff *c, struct tlv *t)
{
	return -1;
}

struct tlv *u_cmdu_reserve_tlv(struct cmdu_buff *c, uint16_t tlv_datalen)
{
	return NULL;
}

uint16_t u_tlv_length(struct tlv *t)
{
	return 0;
}

static void *resolv_function(const char *lib, const char *function)
{
	void *handle;
	void *func = NULL;
	char *error;

	handle = dlopen (lib, RTLD_LAZY);
	if (!handle) {
		fprintf(stderr, "%s: error = %s\n", __func__, dlerror());
		return NULL;
	}

	func = dlsym(handle, function);
	if ((error = dlerror()) != NULL)  {
		fprintf(stderr, "%s\n", error);
		dlclose(handle);
		return NULL;
	}

	dlclose(handle);
	return func;
}

#define EMPTY_FUNC_IF_UNDEF(f)		\
do {					\
	if (!(t_ ## f))			\
		t_ ## f = u_ ## f;	\
} while(0)

int topologyd_use_ieee1905(void)
{
	int ret = 1;


	t_cmdu_ackq_init = resolv_function(IEEE1905_LIB, "cmdu_ackq_init");
	if (!t_cmdu_ackq_init)
		ret = 0;

	t_cmdu_ackq_free = resolv_function(IEEE1905_LIB, "cmdu_ackq_free");
	t_cmdu_ackq_enqueue = resolv_function(IEEE1905_LIB, "cmdu_ackq_enqueue");
	t_cmdu_ackq_dequeue = resolv_function(IEEE1905_LIB, "cmdu_ackq_dequeue");

	t_cmdu_alloc_custom = resolv_function(IEEE1905_LIB, "cmdu_alloc_custom");
	t_cmdu_alloc_simple = resolv_function(IEEE1905_LIB, "cmdu_alloc_simple");
	t_cmdu_free = resolv_function(IEEE1905_LIB, "cmdu_free");
	t_cmdu_get_mid = resolv_function(IEEE1905_LIB, "cmdu_get_mid");
	t_cmdu_get_fid = resolv_function(IEEE1905_LIB, "cmdu_get_fid");
	t_cmdu_get_type = resolv_function(IEEE1905_LIB, "cmdu_get_type");
	t_cmdu_parse_tlvs = resolv_function(IEEE1905_LIB, "cmdu_parse_tlvs");
	t_cmdu_put_eom = resolv_function(IEEE1905_LIB, "cmdu_put_eom");
	t_cmdu_put_tlv = resolv_function(IEEE1905_LIB, "cmdu_put_tlv");
	t_cmdu_reserve_tlv = resolv_function(IEEE1905_LIB, "cmdu_reserve_tlv");
	t_tlv_length = resolv_function(IEEE1905_LIB, "tlv_length");

	t_map_prepare_cmdu_mask = resolv_function(MAP_PLUGIN_LIB, "map_prepare_cmdu_mask");
	t_map_subscribe = resolv_function(MAP_PLUGIN_LIB, "map_subscribe");

	EMPTY_FUNC_IF_UNDEF(cmdu_ackq_init);
	EMPTY_FUNC_IF_UNDEF(cmdu_ackq_free);
	EMPTY_FUNC_IF_UNDEF(cmdu_ackq_enqueue);
	EMPTY_FUNC_IF_UNDEF(cmdu_ackq_dequeue);
	EMPTY_FUNC_IF_UNDEF(cmdu_alloc_custom);
	EMPTY_FUNC_IF_UNDEF(cmdu_alloc_simple);
	EMPTY_FUNC_IF_UNDEF(cmdu_free);
	EMPTY_FUNC_IF_UNDEF(cmdu_get_mid);
	EMPTY_FUNC_IF_UNDEF(cmdu_get_fid);
	EMPTY_FUNC_IF_UNDEF(cmdu_get_type);
	EMPTY_FUNC_IF_UNDEF(cmdu_parse_tlvs);
	EMPTY_FUNC_IF_UNDEF(cmdu_put_eom);
	EMPTY_FUNC_IF_UNDEF(cmdu_put_tlv);
	EMPTY_FUNC_IF_UNDEF(cmdu_reserve_tlv);
	EMPTY_FUNC_IF_UNDEF(tlv_length);

	EMPTY_FUNC_IF_UNDEF(map_prepare_cmdu_mask);
	EMPTY_FUNC_IF_UNDEF(map_subscribe);

	return ret;
}
